@@ -1,6 +1,4 @@
-# -*- coding: utf-8 -*-
-#
-# Copyright 2010-2015 The pygit2 contributors
+# Copyright 2010-2021 The pygit2 contributors
 #
 # This file is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License, version 2,
@@ -25,20 +23,21 @@
 # the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 # Boston, MA 02110-1301, USA.
 
-# Import from the future
-from __future__ import absolute_import, unicode_literals
-
 import weakref
 
 # Import from pygit2
-from _pygit2 import Oid, Tree, Diff
+from ._pygit2 import Oid, Tree, Diff
 from .errors import check_error
 from .ffi import ffi, C
-from .utils import is_string, to_bytes, to_str
+from .utils import to_bytes, to_str
 from .utils import GenericIterator, StrArray
 
 
-class Index(object):
+class Index:
+    # XXX Implement the basic features in C (_pygit2.Index) and make
+    # pygit2.Index to inherit from _pygit2.Index? This would allow for
+    # a proper implementation in some places: e.g. checking the index type
+    # from C code (see Tree_diff_to_index)
 
     def __init__(self, path=None):
         """Create a new Index
@@ -83,12 +82,15 @@ class Index(object):
 
     def __getitem__(self, key):
         centry = ffi.NULL
-        if is_string(key):
+        if isinstance(key, str) or hasattr(key, '__fspath__'):
             centry = C.git_index_get_bypath(self._index, to_bytes(key), 0)
-        elif not key >= 0:
-            raise ValueError(key)
+        elif isinstance(key, int):
+            if key >= 0:
+                centry = C.git_index_get_byindex(self._index, key)
+            else:
+                raise ValueError(key)
         else:
-            centry = C.git_index_get_byindex(self._index, key)
+            raise TypeError('Expected str or int, got %s' % type(key))
 
         if centry == ffi.NULL:
             raise KeyError(key)
@@ -99,23 +101,23 @@ class Index(object):
         return GenericIterator(self)
 
     def read(self, force=True):
-        """Update the contents the Index
+        """
+        Update the contents of the Index by reading from a file.
 
-        Update the contents by reading from a file
+        Parameters:
 
-        Arguments:
-
-        force: if True (the default) allways reload. If False, only if
-        the file has changed
+        force
+            If True (the default) always reload. If False, only if the file
+            has changed.
         """
 
         err = C.git_index_read(self._index, force)
-        check_error(err, True)
+        check_error(err, io=True)
 
     def write(self):
         """Write the contents of the Index to disk."""
         err = C.git_index_write(self._index)
-        check_error(err, True)
+        check_error(err, io=True)
 
     def clear(self):
         err = C.git_index_clear(self._index)
@@ -129,7 +131,7 @@ class Index(object):
         inserted into the Index.
         """
         repo = self._repo
-        if is_string(tree):
+        if isinstance(tree, str):
             tree = repo[tree]
 
         if isinstance(tree, Oid):
@@ -157,6 +159,9 @@ class Index(object):
         It returns the id of the resulting tree.
         """
         coid = ffi.new('git_oid *')
+
+        repo = repo or self._repo
+
         if repo:
             err = C.git_index_write_tree_to(coid, self._index, repo._repo)
         else:
@@ -169,7 +174,14 @@ class Index(object):
         """Remove an entry from the Index.
         """
         err = C.git_index_remove(self._index, to_bytes(path), level)
-        check_error(err, True)
+        check_error(err, io=True)
+
+    def remove_all(self, pathspecs):
+        """Remove all index entries matching pathspecs.
+        """
+        with StrArray(pathspecs) as arr:
+            err = C.git_index_remove_all(self._index, arr, ffi.NULL, ffi.NULL)
+            check_error(err, io=True)
 
     def add_all(self, pathspecs=[]):
         """Add or update index entries matching files in the working directory.
@@ -179,7 +191,7 @@ class Index(object):
         """
         with StrArray(pathspecs) as arr:
             err = C.git_index_add_all(self._index, arr, 0, ffi.NULL, ffi.NULL)
-            check_error(err, True)
+            check_error(err, io=True)
 
     def add(self, path_or_entry):
         """Add or update an entry in the Index.
@@ -191,32 +203,35 @@ class Index(object):
         If an IndexEntry is given, that entry will be added or update in the
         Index without checking for the existence of the path or id.
         """
-
-        if is_string(path_or_entry):
-            path = path_or_entry
-            err = C.git_index_add_bypath(self._index, to_bytes(path))
-        elif isinstance(path_or_entry, IndexEntry):
+        if isinstance(path_or_entry, IndexEntry):
             entry = path_or_entry
             centry, str_ref = entry._to_c()
             err = C.git_index_add(self._index, centry)
+        elif isinstance(path_or_entry, str) or hasattr(path_or_entry, '__fspath__'):
+            path = path_or_entry
+            err = C.git_index_add_bypath(self._index, to_bytes(path))
         else:
-            raise AttributeError('argument must be string or IndexEntry')
+            raise TypeError('argument must be string or IndexEntry')
 
-        check_error(err, True)
+        check_error(err, io=True)
 
     def diff_to_workdir(self, flags=0, context_lines=3, interhunk_lines=0):
-        """Diff the index against the working directory. Return a <Diff> object
+        """
+        Diff the index against the working directory. Return a <Diff> object
         with the differences between the index and the working copy.
 
-        Arguments:
+        Parameters:
 
-        flags: a GIT_DIFF_* constant.
+        flags
+            A GIT_DIFF_* constant.
 
-        context_lines: the number of unchanged lines that define the
-        boundary of a hunk (and to display before and after)
+        context_lines
+            The number of unchanged lines that define the boundary of a hunk
+            (and to display before and after).
 
-        interhunk_lines: the maximum number of unchanged lines between hunk
-        boundaries before the hunks will be merged into a one
+        interhunk_lines
+            The maximum number of unchanged lines between hunk boundaries
+            before the hunks will be merged into a one.
         """
         repo = self._repo
         if repo is None:
@@ -238,20 +253,25 @@ class Index(object):
         return Diff.from_c(bytes(ffi.buffer(cdiff)[:]), repo)
 
     def diff_to_tree(self, tree, flags=0, context_lines=3, interhunk_lines=0):
-        """Diff the index against a tree.  Return a <Diff> object with the
+        """
+        Diff the index against a tree.  Return a <Diff> object with the
         differences between the index and the given tree.
 
-        Arguments:
+        Parameters:
 
-        tree: the tree to diff.
+        tree
+            The tree to diff.
 
-        flags: a GIT_DIFF_* constant.
+        flags
+            A GIT_DIFF_* constant.
 
-        context_lines: the number of unchanged lines that define the boundary
-        of a hunk (and to display before and after)
+        context_lines
+            The number of unchanged lines that define the boundary of a hunk
+            (and to display before and after).
 
-        interhunk_lines: the maximum number of unchanged lines between hunk
-        boundaries before the hunks will be merged into a one.
+        interhunk_lines
+            The maximum number of unchanged lines between hunk boundaries
+            before the hunks will be merged into a one.
         """
         repo = self._repo
         if repo is None:
@@ -292,13 +312,13 @@ class Index(object):
         that represents the conflicts in the index.
 
         This object presents a mapping interface with the paths as keys. You
-        can use the ``del`` operator to remove a conflict form the Index.
+        can use the ``del`` operator to remove a conflict from the Index.
 
         Each conflict is made up of three elements. Access or iteration
         of the conflicts returns a three-tuple of
         :py:class:`~pygit2.IndexEntry`. The first is the common
-        ancestor, the second is the "ours" side of the conflict and the
-        thirs is the "theirs" side.
+        ancestor, the second is the "ours" side of the conflict, and the
+        third is the "theirs" side.
 
         These elements may be None depending on which sides exist for
         the particular conflict.
@@ -315,7 +335,7 @@ class Index(object):
         return self._conflicts()
 
 
-class IndexEntry(object):
+class IndexEntry:
     __slots__ = ['id', 'path', 'mode']
 
     def __init__(self, path, object_id, mode):
@@ -336,6 +356,22 @@ class IndexEntry(object):
         """The id of the referenced object as a hex string"""
         return self.id.hex
 
+    def __str__(self):
+        return "<path={} id={} mode={}>".format(self.path, self.hex, self.mode)
+
+    def __repr__(self):
+        t = type(self)
+        return "<{}.{} path={} id={} mode={}>".format(
+            t.__module__, t.__qualname__, self.path, self.hex, self.mode
+        )
+
+    def __eq__(self, other):
+        if self is other:
+            return True
+        if not isinstance(other, IndexEntry):
+            return NotImplemented
+        return self.path == other.path and self.id == other.id and self.mode == other.mode
+
     def _to_c(self):
         """Convert this entry into the C structure
 
@@ -364,7 +400,7 @@ class IndexEntry(object):
         return entry
 
 
-class ConflictCollection(object):
+class ConflictCollection:
 
     def __init__(self, index):
         self._index = index
@@ -392,7 +428,7 @@ class ConflictCollection(object):
         return ConflictIterator(self._index)
 
 
-class ConflictIterator(object):
+class ConflictIterator:
 
     def __init__(self, index):
         citer = ffi.new('git_index_conflict_iterator **')
@@ -423,3 +459,6 @@ class ConflictIterator(object):
         theirs = IndexEntry._from_c(ctheirs[0])
 
         return ancestor, ours, theirs
+
+    def __iter__(self):
+        return self
